@@ -1,24 +1,27 @@
 # -*- coding: utf-8 -*-
 
-__version__ = "0.4.0"
+__version__ = "0.4.3"
 
 
-import codecs
 import decimal
+import ast
+import operator
 from itertools import product
+from math import sqrt
 from string import ascii_letters
 
-from .simpleeval import SimpleEval
-
 import click
 
+import drang_run.simpleeval as simpleeval
 
-# The arguments for -f and -s come in as raw strings, but we
-# need to be able to interpret things like \t and \n as escape
-# sequences, not literals.
-def interpret(s):
-    if s:
-        return codecs.escape_decode(bytes(s, "utf8"))[0].decode("utf8")
+
+def interpret(text):
+    """
+    Decode text to allow escape sequences (e.g. \n or \t) input by the user
+    to be processed as expected. Important for the -f and -s options.
+    """
+    if text:
+        return bytes(text, "utf-8").decode("unicode_escape")
     else:
         return None
 
@@ -80,6 +83,10 @@ class Counter(object):
 
         if reverse:
             self.start, self.stop = self.stop, self.start
+            # if the original stop is not in the run, the reversed
+            # run should start at the last value of the not reversed run
+            if ((self.start - self.stop) % self.step) != 0:
+                self.start -= (self.start - self.stop) % self.step
             self.step = -1 * self.step
             self.reversed = not self.reversed
 
@@ -124,6 +131,7 @@ class OptargCommand(click.Command):
     metavar="START STOP STEP",
 )
 @click.option(
+    "-d",
     "--def",
     "var_defs",
     multiple=True,
@@ -131,11 +139,12 @@ class OptargCommand(click.Command):
     help="Define a variable to use in the format string. The Variable may contain simple arithmetic expressions.",
     metavar="NAME EXPR",
 )
+@click.option("-F", "--filter", help="An boolean expression to filter out some values.", metavar="EXPR")
 @click.argument("start", required=False, default="1")
 @click.argument("stop")
 @click.argument("step", required=False, default="1")
 @click.pass_context
-def run(ctx, start, stop, step, fstring, sep, reverse, also, var_defs):
+def run(ctx, start, stop, step, fstring, sep, reverse, also, var_defs, filter):
     """Generate a run of integers or characters. Similar to jot and seq.
 
     The run of numbers can be integers or reals, depending on the values of START, STOP, and STEP.
@@ -156,35 +165,63 @@ def run(ctx, start, stop, step, fstring, sep, reverse, also, var_defs):
             except ValueError as err:
                 ctx.fail(str(err))
 
+    # Initialize variable defintions
     variables = {name: 0 for name, ex in var_defs}
 
-    run = list(product(*counters))
+    # Lazy evaluation of counter product
+    # Each iteration is printed directly after evaluation.
+    # itertools.product still takes a moment to build the cartesian
+    # product of counter values for large inputs.
 
-    # Convert to text
-    try:
-        # Create restricted evaluator for --def params
-        simple = SimpleEval()
-        # simple.functions = {}
+    # Create restricted evaluator singleton
+    simple = simpleeval.SimpleEval()
+    simple.functions.update(sqrt=sqrt)
+    simple.operators[ast.BitOr] = operator.or_
+    simple.operators[ast.BitAnd] = operator.and_
+    simple.operators[ast.BitXor] = operator.xor
 
-        runText = []
-        for numbers in run:
-            for name, ex in var_defs:
-                for i, v in enumerate(numbers):
-                    ex = ex.replace(f"{{{i}}}", str(v))
-                for n, v in variables.items():
-                    ex = ex.replace(f"{{{n}}}", str(v))
-                variables[name] = simple.eval(ex)
-            runText.append(fstring.format(*numbers, **variables))
-        # runText = [fstring.format(*n, **variables) for n in run]
-    except Exception as err:
-        format_opt = next(
-            a for a in ctx.command.params if a.human_readable_name == "fstring"
-        )
-        hint = format_opt.get_error_hint(ctx)
-        ctx.fail(f"Invalid value for {hint}: {str(err)}")
+    is_first = True
+    for numbers in product(*counters):
+        try:
+            # Evaluate expressions in variable definitions
+            # for this instance of the counters
+            for name, expression in var_defs:
+                expression_ = expression.format(*numbers, **variables)
+                variables[name] = simple.eval(expression_)
+        except Exception as err:
+            format_opt = next(
+                a for a in ctx.command.params if a.human_readable_name == "var_defs"
+            )
+            hint = format_opt.get_error_hint(ctx)
+            ctx.fail(f"Invalid value for {hint}: {str(err)}")
 
-    # Reverse the list if asked.
-    # if reverse:
-    #    runText.reverse()
+        try:
+            # Evaluate filter expression if provided
+            # Use counter values and variable definitions after evaluation
+            if filter is not None:
+                filter_ = filter.format(*numbers, **variables)
+                filter_ = simple.eval(filter_)
+                if not bool(filter_):
+                    continue
+        except Exception as err:
+            format_opt = next(
+                a for a in ctx.command.params if a.human_readable_name == "filter"
+            )
+            hint = format_opt.get_error_hint(ctx)
+            ctx.fail(f"Invalid value for {hint}: {str(err)}")
 
-    click.echo(sep.join(runText))
+        try:
+            # Convert iteration to text, using the provided format string
+            if not is_first:
+                click.echo(sep, nl=False)
+            else:
+                is_first = False
+            click.echo(fstring.format(*numbers, **variables), nl=False)
+        except Exception as err:
+            format_opt = next(
+                a for a in ctx.command.params if a.human_readable_name == "fstring"
+            )
+            hint = format_opt.get_error_hint(ctx)
+            ctx.fail(f"Invalid value for {hint}: {str(err)}")
+    # Print final newline
+    click.echo("\n", nl=False)
